# Copyright 2017-2024 Lawrence Livermore National Security, LLC and other
# Hatchet Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

dot_keywords = ["graph", "subgraph", "digraph", "node", "edge", "strict"]
